""""N2V Algorithm configuration."""

from typing import Annotated, Literal

from pydantic import AfterValidator

from careamics.config.architectures import UNetModel
from careamics.config.validators import (
    model_matching_in_out_channels,
    model_without_final_activation,
)

from .unet_algorithm_model import UNetBasedAlgorithm


class N2VAlgorithm(UNetBasedAlgorithm):
    """N2V Algorithm configuration."""

    algorithm: Literal["n2v"] = "n2v"
    """N2V Algorithm name."""

    loss: Literal["n2v"] = "n2v"
    """N2V loss function."""

    model: Annotated[
        UNetModel,
        AfterValidator(model_matching_in_out_channels),
        AfterValidator(model_without_final_activation),
    ]



import pytest

from careamics.config.data import N2VDataConfig
from careamics.config.support import (
    SupportedPixelManipulation,
    SupportedStructAxis,
    SupportedTransform,
)


def test_error_no_manipulate(minimum_data: dict):
    """Test that an error is raised if no N2VManipulate transform is passed."""
    minimum_data["transforms"] = [
        {"name": SupportedTransform.XY_FLIP.value},
        {"name": SupportedTransform.XY_RANDOM_ROTATE90.value},
    ]
    with pytest.raises(ValueError):
        N2VDataConfig(**minimum_data)


@pytest.mark.parametrize(
    "transforms",
    [
        [
            {"name": SupportedTransform.N2V_MANIPULATE.value},
            {"name": SupportedTransform.XY_FLIP.value},
        ],
        [
            {"name": SupportedTransform.XY_FLIP.value},
            {"name": SupportedTransform.N2V_MANIPULATE.value},
            {"name": SupportedTransform.XY_RANDOM_ROTATE90.value},
        ],
    ],
)
def test_n2vmanipulate_not_last_transform(minimum_data: dict, transforms):
    """Test that N2V Manipulate not in the last position raises an error."""
    minimum_data["transforms"] = transforms
    with pytest.raises(ValueError):
        N2VDataConfig(**minimum_data)


def test_multiple_n2v_manipulate(minimum_data: dict):
    """Test that passing multiple n2v manipulate raises an error."""
    minimum_data["transforms"] = [
        {"name": SupportedTransform.N2V_MANIPULATE.value},
        {"name": SupportedTransform.N2V_MANIPULATE.value},
    ]
    with pytest.raises(ValueError):
        N2VDataConfig(**minimum_data)


def test_correct_transform_parameters(minimum_data: dict):
    """Test that the transforms have the correct parameters.

    This is important to know that the transforms are not all instantiated as
    a generic transform.
    """
    minimum_data["transforms"] = [
        {"name": SupportedTransform.XY_FLIP.value},
        {"name": SupportedTransform.XY_RANDOM_ROTATE90.value},
        {"name": SupportedTransform.N2V_MANIPULATE.value},
    ]
    model = N2VDataConfig(**minimum_data)

    # N2VManipulate
    params = model.transforms[-1].model_dump()
    assert "roi_size" in params
    assert "masked_pixel_percentage" in params
    assert "strategy" in params
    assert "struct_mask_axis" in params
    assert "struct_mask_span" in params


def test_set_n2v_strategy(minimum_data: dict):
    """Test that the N2V strategy can be set."""
    uniform = SupportedPixelManipulation.UNIFORM.value
    median = SupportedPixelManipulation.MEDIAN.value

    data = N2VDataConfig(**minimum_data)
    assert data.transforms[-1].name == SupportedTransform.N2V_MANIPULATE.value
    assert data.transforms[-1].strategy == uniform

    data.set_masking_strategy(median)
    assert data.transforms[-1].strategy == median

    data.set_masking_strategy(uniform)
    assert data.transforms[-1].strategy == uniform


def test_set_n2v_strategy_wrong_value(minimum_data: dict):
    """Test that passing a wrong strategy raises an error."""
    data = N2VDataConfig(**minimum_data)
    with pytest.raises(ValueError):
        data.set_masking_strategy("wrong_value")


def test_set_struct_mask(minimum_data: dict):
    """Test that the struct mask can be set."""
    none = SupportedStructAxis.NONE.value
    vertical = SupportedStructAxis.VERTICAL.value
    horizontal = SupportedStructAxis.HORIZONTAL.value

    data = N2VDataConfig(**minimum_data)
    assert data.transforms[-1].name == SupportedTransform.N2V_MANIPULATE.value
    assert data.transforms[-1].struct_mask_axis == none
    assert data.transforms[-1].struct_mask_span == 5

    data.set_structN2V_mask(vertical, 3)
    assert data.transforms[-1].struct_mask_axis == vertical
    assert data.transforms[-1].struct_mask_span == 3

    data.set_structN2V_mask(horizontal, 7)
    assert data.transforms[-1].struct_mask_axis == horizontal
    assert data.transforms[-1].struct_mask_span == 7

    data.set_structN2V_mask(none, 11)
    assert data.transforms[-1].struct_mask_axis == none
    assert data.transforms[-1].struct_mask_span == 11


def test_set_struct_mask_wrong_value(minimum_data: dict):
    """Test that passing a wrong struct mask axis raises an error."""
    data = N2VDataConfig(**minimum_data)
    with pytest.raises(ValueError):
        data.set_structN2V_mask("wrong_value", 3)

    with pytest.raises(ValueError):
        data.set_structN2V_mask(SupportedStructAxis.VERTICAL.value, 1)



import pytest

from careamics.config import (
    N2VConfiguration,
)
from careamics.config.support import (
    SupportedAlgorithm,
    SupportedPixelManipulation,
)


@pytest.mark.parametrize(
    "algorithm, strategy",
    [
        ("n2v", SupportedPixelManipulation.UNIFORM.value),
        ("n2v2", SupportedPixelManipulation.MEDIAN.value),
    ],
)
def test_correct_n2v2_and_transforms(
    minimum_n2v_configuration: dict, algorithm, strategy
):
    """Test that N2V and N2V2 are correctly instantiated."""
    minimum_n2v_configuration["algorithm_config"] = {
        "algorithm": "n2v",
        "loss": "n2v",
        "model": {
            "architecture": "UNet",
            "n2v2": algorithm == "n2v2",
        },
    }
    minimum_n2v_configuration["data_config"]["transforms"] = [
        {
            "name": "N2VManipulate",
            "strategy": strategy,
        }
    ]

    N2VConfiguration(**minimum_n2v_configuration)


@pytest.mark.parametrize(
    "algorithm, strategy",
    [
        ("n2v", SupportedPixelManipulation.MEDIAN.value),
        ("n2v2", SupportedPixelManipulation.UNIFORM.value),
    ],
)
def test_wrong_n2v2_and_transforms(
    minimum_n2v_configuration: dict, algorithm, strategy
):
    """Test that N2V and N2V2 throw an error if the strategy and the N2V2 UNet
    parameters disagree."""
    minimum_n2v_configuration["algorithm_config"] = {
        "algorithm": "n2v",
        "loss": "n2v",
        "model": {
            "architecture": "UNet",
            "n2v2": algorithm == "n2v2",
        },
    }
    minimum_n2v_configuration["data_config"]["transforms"] = [
        {
            "name": "N2VManipulate",
            "strategy": strategy,
        }
    ]

    with pytest.raises(ValueError):
        N2VConfiguration(**minimum_n2v_configuration)


def test_setting_n2v2(minimum_n2v_configuration: dict):
    # make sure we use n2v
    minimum_n2v_configuration["algorithm_config"][
        "algorithm"
    ] = SupportedAlgorithm.N2V.value

    # test config
    config = N2VConfiguration(**minimum_n2v_configuration)
    assert config.algorithm_config.algorithm == SupportedAlgorithm.N2V.value
    assert not config.algorithm_config.model.n2v2
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.UNIFORM.value
    )

    # set N2V2
    config.set_n2v2(True)
    assert config.algorithm_config.model.n2v2
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.MEDIAN.value
    )

    # set back to N2V
    config.set_n2v2(False)
    assert not config.algorithm_config.model.n2v2
    assert (
        config.data_config.transforms[-1].strategy
        == SupportedPixelManipulation.UNIFORM.value
    )



"""Noise2Void specific data configuration model."""

from collections.abc import Sequence
from typing import Literal

from pydantic import Field, field_validator

from careamics.config.data.data_model import GeneralDataConfig
from careamics.config.support import SupportedTransform
from careamics.config.transformations import (
    N2V_TRANSFORMS_UNION,
    N2VManipulateModel,
    XYFlipModel,
    XYRandomRotate90Model,
)


class N2VDataConfig(GeneralDataConfig):
    """N2V specific data configuration model."""

    transforms: Sequence[N2V_TRANSFORMS_UNION] = Field(
        default=[XYFlipModel(), XYRandomRotate90Model(), N2VManipulateModel()],
        validate_default=True,
    )
    """N2V compatible transforms. N2VManpulate should be the last transform."""

    @field_validator("transforms")
    @classmethod
    def validate_transforms(
        cls, transforms: list[N2V_TRANSFORMS_UNION]
    ) -> list[N2V_TRANSFORMS_UNION]:
        """
        Validate N2VManipulate transform position in the transform list.

        Parameters
        ----------
        transforms : list of transforms compatible with N2V
            Transforms.

        Returns
        -------
        list of transforms
            Validated transforms.

        Raises
        ------
        ValueError
            If multiple instances of N2VManipulate are found or if it is not the last
            transform.
        """
        transform_list = [t.name for t in transforms]

        if SupportedTransform.N2V_MANIPULATE in transform_list:
            # multiple N2V_MANIPULATE
            if transform_list.count(SupportedTransform.N2V_MANIPULATE.value) > 1:
                raise ValueError(
                    f"Multiple instances of "
                    f"{SupportedTransform.N2V_MANIPULATE} transforms "
                    f"are not allowed."
                )

            # N2V_MANIPULATE not the last transform
            elif transform_list[-1] != SupportedTransform.N2V_MANIPULATE:
                raise ValueError(
                    f"{SupportedTransform.N2V_MANIPULATE} transform "
                    f"should be the last transform."
                )

        else:
            raise ValueError(
                f"{SupportedTransform.N2V_MANIPULATE} transform "
                f"is required for N2V training."
            )

        return transforms

    def set_n2v2(self, use_n2v2: bool) -> None:
        """
        Set the N2V transform to the N2V2 version.

        Parameters
        ----------
        use_n2v2 : bool
            Whether to use N2V2.

        Raises
        ------
        ValueError
            If the N2V pixel manipulate transform is not found in the transforms.
        """
        if use_n2v2:
            self.set_masking_strategy("median")
        else:
            self.set_masking_strategy("uniform")

    def set_masking_strategy(self, strategy: Literal["uniform", "median"]) -> None:
        """
        Set masking strategy.

        Parameters
        ----------
        strategy : "uniform" or "median"
            Strategy to use for N2V2.

        Raises
        ------
        ValueError
            If the N2V pixel manipulate transform is not found in the transforms.
        """
        found_n2v = False

        for transform in self.transforms:
            if transform.name == SupportedTransform.N2V_MANIPULATE.value:
                transform.strategy = strategy
                found_n2v = True

        if not found_n2v:
            transforms = [t.name for t in self.transforms]
            raise ValueError(
                f"N2V_Manipulate transform not found in the transforms list "
                f"({transforms})."
            )

    def get_masking_strategy(self) -> Literal["uniform", "median"]:
        """
        Get N2V2 strategy.

        Returns
        -------
        "uniform" or "median"
            Strategy used for N2V2.
        """
        for transform in self.transforms:
            if transform.name == SupportedTransform.N2V_MANIPULATE.value:
                return transform.strategy

        raise ValueError(
            f"{SupportedTransform.N2V_MANIPULATE} transform "
            f"is required for N2V training."
        )

    def set_structN2V_mask(
        self, mask_axis: Literal["horizontal", "vertical", "none"], mask_span: int
    ) -> None:
        """
        Set structN2V mask parameters.

        Setting `mask_axis` to `none` will disable structN2V.

        Parameters
        ----------
        mask_axis : Literal["horizontal", "vertical", "none"]
            Axis along which to apply the mask. `none` will disable structN2V.
        mask_span : int
            Total span of the mask in pixels.

        Raises
        ------
        ValueError
            If the N2V pixel manipulate transform is not found in the transforms.
        """
        found_n2v = False

        for transform in self.transforms:
            if transform.name == SupportedTransform.N2V_MANIPULATE.value:
                transform.struct_mask_axis = mask_axis
                transform.struct_mask_span = mask_span
                found_n2v = True

        if not found_n2v:
            transforms = [t.name for t in self.transforms]
            raise ValueError(
                f"N2V pixel manipulate transform not found in the transforms "
                f"({transforms})."
            )

    def is_using_struct_n2v(self) -> bool:
        """
        Check if structN2V is enabled.

        Returns
        -------
        bool
            Whether structN2V is enabled or not.
        """
        for transform in self.transforms:
            if transform.name == SupportedTransform.N2V_MANIPULATE.value:
                return transform.struct_mask_axis != "none"

        raise ValueError(
            f"N2V pixel manipulate transform not found in the transforms "
            f"({self.transforms})."
        )



"""Data Pydantic configuration models."""

__all__ = [
    "DataConfig",
    "GeneralDataConfig",
    "N2VDataConfig",
]

from .data_model import DataConfig, GeneralDataConfig
from .n2v_data_model import N2VDataConfig



"""N2N Algorithm configuration."""

from typing import Annotated, Literal

from pydantic import AfterValidator

from careamics.config.architectures import UNetModel
from careamics.config.validators import (
    model_without_final_activation,
    model_without_n2v2,
)

from .unet_algorithm_model import UNetBasedAlgorithm


class N2NAlgorithm(UNetBasedAlgorithm):
    """Noise2Noise Algorithm configuration."""

    algorithm: Literal["n2n"] = "n2n"
    """N2N Algorithm name."""

    loss: Literal["mae", "mse"] = "mae"
    """N2N-compatible loss function."""

    model: Annotated[
        UNetModel,
        AfterValidator(model_without_n2v2),
        AfterValidator(model_without_final_activation),
    ]
    """UNet without a final activation function and without the `n2v2` modifications."""
